"""Client provides an easy-to use client library to connect to Axiom."""

import ndjson
import dacite
import gzip
import ujson
import os
from .util import Util
from enum import Enum
from humps import decamelize
from typing import Optional, List, Dict, Any
from logging import getLogger
from dataclasses import dataclass, field, asdict
from datetime import datetime
from requests_toolbelt.sessions import BaseUrlSession
from requests_toolbelt.utils.dump import dump_response
from requests.adapters import HTTPAdapter, Retry
from .datasets import DatasetsClient
from .query import QueryLegacy, QueryResult, QueryOptions, QueryLegacyResult, QueryKind
from .annotations import AnnotationsClient
from .users import UsersClient
from .__init__ import __version__


AXIOM_URL = "https://api.axiom.co"


@dataclass
class Error:
    status: Optional[int] = field(default=None)
    message: Optional[str] = field(default=None)
    error: Optional[str] = field(default=None)


@dataclass
class IngestFailure:
    """The ingestion failure of a single event"""

    timestamp: datetime
    error: str


@dataclass
class IngestStatus:
    """The status after an event ingestion operation"""

    ingested: int
    failed: int
    failures: List[IngestFailure]
    processed_bytes: int
    blocks_created: int
    wal_length: int


@dataclass
class IngestOptions:
    """IngestOptions specifies the optional parameters for the Ingest and
    IngestEvents method of the Datasets service."""

    # timestamp field defines a custom field to extract the ingestion timestamp
    # from. Defaults to `_time`.
    timestamp_field: str = field(default="_time")
    # timestamp format defines a custom format for the TimestampField.
    # The reference time is `Mon Jan 2 15:04:05 -0700 MST 2006`, as specified
    # in https://pkg.go.dev/time/?tab=doc#Parse.
    timestamp_format: Optional[str] = field(default=None)
    # CSV delimiter is the delimiter that separates CSV fields. Only valid when
    # the content to be ingested is CSV formatted.
    CSV_delimiter: Optional[str] = field(default=None)


class AplResultFormat(Enum):
    """The result format of an APL query."""

    Legacy = "legacy"


class ContentType(Enum):
    """ContentType describes the content type of the data to ingest."""

    JSON = "application/json"
    NDJSON = "application/x-ndjson"
    CSV = "text/csv"


class ContentEncoding(Enum):
    """ContentEncoding describes the content encoding of the data to ingest."""

    IDENTITY = "1"
    GZIP = "gzip"


class WrongQueryKindException(Exception):
    pass


@dataclass
class AplOptions:
    """AplOptions specifies the optional parameters for the apl query method."""

    start_time: Optional[datetime] = field(default=None)
    end_time: Optional[datetime] = field(default=None)
    no_cache: bool = field(default=False)
    save: bool = field(default=False)
    format: AplResultFormat = field(default=AplResultFormat.Legacy)


def raise_response_error(r):
    if r.status_code >= 400:
        print("==== Response Debugging ====")
        print("##Request Headers", r.request.headers)

        # extract content type
        ct = r.headers["content-type"].split(";")[0]
        if ct == ContentType.JSON.value:
            dump = dump_response(r)
            print(dump)
            print("##Response:", dump.decode("UTF-8"))
            err = dacite.from_dict(data_class=Error, data=r.json())
            print(err)
        elif ct == ContentType.NDJSON.value:
            decoded = ndjson.loads(r.text)
            print("##Response:", decoded)

        r.raise_for_status()
        # TODO: Decode JSON https://github.com/axiomhq/axiom-go/blob/610cfbd235d3df17f96a4bb156c50385cfbd9edd/axiom/error.go#L35-L50


class Client:  # pylint: disable=R0903
    """The client class allows you to connect to Axiom."""

    datasets: DatasetsClient
    users: UsersClient
    annotations: AnnotationsClient

    def __init__(
        self,
        token: Optional[str] = None,
        org_id: Optional[str] = None,
        url_base: Optional[str] = None,
    ):
        # fallback to env variables if token, org_id or url are not provided
        if token is None:
            token = os.getenv("AXIOM_TOKEN")
        if org_id is None:
            org_id = os.getenv("AXIOM_ORG_ID")
        if url_base is None:
            url_base = AXIOM_URL

        self.logger = getLogger()
        # set exponential retries
        retries = Retry(
            total=3, backoff_factor=2, status_forcelist=[500, 502, 503, 504]
        )

        self.session = BaseUrlSession(url_base.rstrip("/"))
        self.session.mount("http://", HTTPAdapter(max_retries=retries))
        self.session.mount("https://", HTTPAdapter(max_retries=retries))
        # hook on responses, raise error when response is not successfull
        self.session.hooks = {
            "response": lambda r, *args, **kwargs: raise_response_error(r)
        }
        self.session.headers.update(
            {
                "Authorization": "Bearer %s" % token,
                # set a default Content-Type header, can be overriden by requests.
                "Content-Type": "application/json",
                "User-Agent": f"axiom-py/{__version__}",
            }
        )

        # if there is an organization id passed,
        # set it in the header
        if org_id:
            self.logger.info("found organization id: %s" % org_id)
            self.session.headers.update({"X-Axiom-Org-Id": org_id})

        self.datasets = DatasetsClient(self.session, self.logger)
        self.users = UsersClient(self.session)
        self.annotations = AnnotationsClient(self.session, self.logger)

    def ingest(
        self,
        dataset: str,
        payload: bytes,
        contentType: ContentType,
        enc: ContentEncoding,
        opts: Optional[IngestOptions] = None,
    ) -> IngestStatus:
        """Ingest the events into the named dataset and returns the status."""
        path = "/v1/datasets/%s/ingest" % dataset

        # check if passed content type and encoding are correct
        if not contentType:
            raise ValueError("unknown content-type, choose one of json,x-ndjson or csv")

        if not enc:
            raise ValueError("unknown content-encoding")

        # set headers
        headers = {"Content-Type": contentType.value, "Content-Encoding": enc.value}
        # prepare query params
        params = self._prepare_ingest_options(opts)

        # override the default header and set the value from the passed parameter
        res = self.session.post(path, data=payload, headers=headers, params=params)
        status_snake = decamelize(res.json())
        return Util.from_dict(IngestStatus, status_snake)

    def ingest_events(
        self,
        dataset: str,
        events: List[dict],
        opts: Optional[IngestOptions] = None,
    ) -> IngestStatus:
        """Ingest the events into the named dataset and returns the status."""
        # encode request payload to NDJSON
        content = ndjson.dumps(events, default=Util.handle_json_serialization).encode(
            "UTF-8"
        )
        gzipped = gzip.compress(content)

        return self.ingest(
            dataset, gzipped, ContentType.NDJSON, ContentEncoding.GZIP, opts
        )

    def query_legacy(
        self, id: str, query: QueryLegacy, opts: QueryOptions
    ) -> QueryLegacyResult:
        """Executes the given query on the dataset identified by its id."""
        if not opts.saveAsKind or (opts.saveAsKind == QueryKind.APL):
            raise WrongQueryKindException(
                "invalid query kind %s: must be %s or %s"
                % (opts.saveAsKind, QueryKind.ANALYTICS, QueryKind.STREAM)
            )

        path = "/v1/datasets/%s/query" % id
        payload = ujson.dumps(asdict(query), default=Util.handle_json_serialization)
        self.logger.debug("sending query %s" % payload)
        params = self._prepare_query_options(opts)
        res = self.session.post(path, data=payload, params=params)
        result = Util.from_dict(QueryLegacyResult, res.json())
        self.logger.debug(f"query result: {result}")
        query_id = res.headers.get("X-Axiom-History-Query-Id")
        self.logger.info(f"received query result with query_id: {query_id}")
        result.savedQueryID = query_id
        return result

    def apl_query(self, apl: str, opts: Optional[AplOptions] = None) -> QueryResult:
        """Executes the given apl query on the dataset identified by its id."""
        return self.query(apl, opts)

    def query(self, apl: str, opts: Optional[AplOptions] = None) -> QueryResult:
        """Executes the given apl query on the dataset identified by its id."""
        path = "/v1/datasets/_apl"
        payload = ujson.dumps(
            self._prepare_apl_payload(apl, opts),
            default=Util.handle_json_serialization,
        )
        self.logger.debug("sending query %s" % payload)
        params = self._prepare_apl_options(opts)
        res = self.session.post(path, data=payload, params=params)
        result = Util.from_dict(QueryResult, res.json())
        self.logger.debug(f"apl query result: {result}")
        query_id = res.headers.get("X-Axiom-History-Query-Id")
        self.logger.info(f"received query result with query_id: {query_id}")
        result.savedQueryID = query_id
        return result

    def _prepare_query_options(self, opts: QueryOptions) -> Dict[str, Any]:
        """returns the query options as a Dict, handles any renaming for key fields."""
        if opts is None:
            return {}
        params = {}
        if opts.streamingDuration:
            params["streaming-duration"] = (
                opts.streamingDuration.seconds.__str__() + "s"
            )
        if opts.saveAsKind:
            params["saveAsKind"] = opts.saveAsKind.value

        params["nocache"] = opts.nocache.__str__()

        return params

    def _prepare_ingest_options(self, opts: Optional[IngestOptions]) -> Dict[str, Any]:
        """the query params for ingest api are expected in a format
        that couldn't be defined as a variable name because it has a dash.
        As a work around, we create the params dict manually."""

        if opts is None:
            return {}

        params = {}
        if opts.timestamp_field:
            params["timestamp-field"] = opts.timestamp_field
        if opts.timestamp_format:
            params["timestamp-format"] = opts.timestamp_format
        if opts.CSV_delimiter:
            params["csv-delimiter"] = opts.CSV_delimiter

        return params

    def _prepare_apl_options(self, opts: Optional[AplOptions]) -> Dict[str, Any]:
        """Prepare the apl query options for the request."""
        params = {}

        if opts is None:
            params["format"] = AplResultFormat.Legacy.value
            return params

        if opts.no_cache:
            params["nocache"] = opts.no_cache.__str__()
        if opts.save:
            params["save"] = opts.save
        if opts.format:
            params["format"] = opts.format.value

        return params

    def _prepare_apl_payload(
        self, apl: str, opts: Optional[AplOptions]
    ) -> Dict[str, Any]:
        """Prepare the apl query options for the request."""
        params = {}
        params["apl"] = apl

        if opts is not None:
            if opts.start_time:
                params["startTime"] = opts.start_time
            if opts.end_time:
                params["endTime"] = opts.end_time

        return params



from typing import List
from .util import Util
from dataclasses import dataclass
from requests import Session


@dataclass
class User:
    """An authenticated axiom user."""

    id: str
    name: str
    emails: List[str]


class UsersClient:
    """The UsersClient is a client for the Axiom Users service."""

    def __init__(self, session: Session):
        self.session = session

    def current(self) -> User:
        """Get the current authenticated user."""
        res = self.session.get("/v1/user")
        user = Util.from_dict(User, res.json())
        return user



from enum import Enum
from typing import List, Optional
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from .aggregation import Aggregation
from .filter import Filter


class QueryKind(Enum):
    """represents kind of query"""

    ANALYTICS = "analytics"
    STREAM = "stream"
    APL = "apl"


@dataclass
class Order:
    """Order specifies the order a queries result will be in."""

    # Field to order on. Must be present in `GroupBy` or used by an
    # aggregation.
    field: str
    # Desc specifies if the field is ordered ascending or descending.
    desc: bool


@dataclass
class VirtualField:
    """
    A VirtualField is not part of a dataset and its value is derived from an
    expression. Aggregations, filters and orders can reference this field like
    any other field.
    """

    # Alias the virtual field is referenced by.
    alias: str
    # Expression which specifies the virtual fields value.
    expr: str


@dataclass
class Projection:
    """A Projection is a field that is projected to the query result."""

    # Field to project to the query result.
    field: str
    # Alias to reference the projected field by. Optional.
    alias: str


@dataclass
class QueryLegacy:
    """represents a query that gets executed on a dataset."""

    # start time of the query. Required.
    startTime: datetime
    # end time of the query. Required.
    endTime: datetime
    # resolution of the queries graph. Valid values are the queries time
    # range / 100 at maximum and / 1000 at minimum. Use zero value for
    # serve-side auto-detection.
    resolution: str = field(default="auto")
    # Aggregations performed as part of the query.
    aggregations: Optional[List[Aggregation]] = field(default_factory=lambda: [])
    # GroupBy is a list of field names to group the query result by. Only valid
    # when at least one aggregation is specified.
    groupBy: Optional[List[str]] = field(default_factory=lambda: [])
    # Filter applied on the queried results.
    filter: Optional[Filter] = field(default=None)
    # Order is a list of order rules that specify the order of the query
    # result.
    order: Optional[List[Order]] = field(default_factory=lambda: [])
    # Limit the amount of results returned from the query.
    limit: int = field(default=10)
    # VirtualFields is a list of virtual fields that can be referenced by
    # aggregations, filters and orders.
    virtualFields: Optional[List[VirtualField]] = field(default_factory=lambda: [])
    # Projections is a list of projections that can be referenced by
    # aggregations, filters and orders. Leaving it empty projects all available
    # fields to the query result.
    project: Optional[List[Projection]] = field(default_factory=lambda: [])
    # Cursor is the query cursor. It should be set to the Cursor returned with
    # a previous query result if it was partial.
    cursor: str = field(default=None)
    # IncludeCursor will return the Cursor as part of the query result, if set
    # to true.
    includeCursor: bool = field(default=False)
    # ContinuationToken is used to get more results of a previous query. It is
    # not valid for starred queries or otherwise stored queries.
    continuationToken: str = field(default=None)



"""Logging contains the AxiomHandler and related methods to do with logging."""

import time
import atexit

from logging import Handler, NOTSET, getLogger, WARNING
from .client import Client


class AxiomHandler(Handler):
    """A logging handler that sends logs to Axiom."""

    client: Client
    dataset: str
    logcache: list
    interval: int
    last_run: float

    def __init__(self, client: Client, dataset: str, level=NOTSET, interval=1):
        Handler.__init__(self, level)
        # set urllib3 logging level to warning, check:
        # https://github.com/axiomhq/axiom-py/issues/23
        # This is a temp solution that would stop requests
        # library from flooding the logs with debug messages
        getLogger("urllib3").setLevel(WARNING)
        self.client = client
        self.dataset = dataset
        self.buffer = []
        self.last_run = time.monotonic()
        self.interval = interval

        # register flush on exit,
        atexit.register(self.flush)

    def emit(self, record):
        """emit sends a log to Axiom."""
        self.buffer.append(record.__dict__)
        if len(self.buffer) >= 1000 or time.monotonic() - self.last_run > self.interval:
            self.flush()

    def flush(self):
        """flush sends all logs in the logcache to Axiom."""
        self.last_run = time.monotonic()
        if len(self.buffer) == 0:
            return
        self.client.ingest_events(self.dataset, self.buffer)
        self.buffer = []



"""This package provides dataset models and methods as well as a DatasetClient"""

import ujson
from logging import Logger
from requests import Session
from typing import List, Dict
from dataclasses import dataclass, asdict, field
from datetime import datetime, timedelta
from .util import Util


@dataclass
class Dataset:
    """Represents an Axiom dataset"""

    id: str = field(init=False)
    name: str
    description: str
    who: str
    created: str


@dataclass
class DatasetCreateRequest:
    """Request used to create a dataset"""

    name: str
    description: str


@dataclass
class DatasetUpdateRequest:
    """Request used to update a dataset"""

    description: str


@dataclass
class TrimRequest:
    """MaxDuration marks the oldest timestamp an event can have before getting deleted."""

    maxDuration: str


@dataclass
class Field:
    """A field of a dataset"""

    name: str
    description: str
    type: str
    unit: str
    hidden: bool


@dataclass
class DatasetInfo:
    """Information and statistics stored inside a dataset"""

    name: str
    numBlocks: int
    numEvents: int
    numFields: int
    inputBytes: int
    inputBytesHuman: str
    compressedBytes: int
    compressedBytesHuman: str
    minTime: datetime
    maxTime: datetime
    fields: List[Field]
    who: str
    created: datetime


class DatasetsClient:  # pylint: disable=R0903
    """DatasetsClient has methods to manipulate datasets."""

    session: Session

    def __init__(self, session: Session, logger: Logger):
        self.session = session
        self.logger = logger

    def get(self, id: str) -> Dataset:
        """Get a dataset by id."""
        path = "/v1/datasets/%s" % id
        res = self.session.get(path)
        decoded_response = res.json()
        return Util.from_dict(Dataset, decoded_response)

    def create(self, req: DatasetCreateRequest) -> Dataset:
        """Create a dataset with the given properties."""
        path = "/v1/datasets"
        res = self.session.post(path, data=ujson.dumps(asdict(req)))
        ds = Util.from_dict(Dataset, res.json())
        self.logger.info(f"created new dataset: {ds.name}")
        return ds

    def get_list(self) -> List[Dataset]:
        """List all available datasets."""
        path = "/v1/datasets"
        res = self.session.get(path)

        datasets = []
        for record in res.json():
            ds = Util.from_dict(Dataset, record)
            datasets.append(ds)

        return datasets

    def update(self, id: str, req: DatasetUpdateRequest) -> Dataset:
        """Update a dataset with the given properties."""
        path = "/v1/datasets/%s" % id
        res = self.session.put(path, data=ujson.dumps(asdict(req)))
        ds = Util.from_dict(Dataset, res.json())
        self.logger.info(f"updated dataset({ds.name}) with new desc: {ds.description}")
        return ds

    def delete(self, id: str):
        """Deletes a dataset with the given id."""
        path = "/v1/datasets/%s" % id
        self.session.delete(path)

    def trim(self, id: str, maxDuration: timedelta):
        """
        Trim the dataset identified by its id to a given length. The max duration
        given will mark the oldest timestamp an event can have. Older ones will be
        deleted from the dataset.
        """
        path = "/v1/datasets/%s/trim" % id
        # prepare request payload and format masDuration to append time unit at the end, e.g `1s`
        req = TrimRequest(f"{maxDuration.seconds}s")
        self.session.post(path, data=ujson.dumps(asdict(req)))

    def info(self, id: str) -> DatasetInfo:
        """Returns the info about a dataset."""
        path = "/v1/datasets/%s/info" % id
        res = self.session.get(path)
        decoded_response = res.json()
        return Util.from_dict(DatasetInfo, decoded_response)



from enum import Enum
from dataclasses import dataclass, field as dataclass_field


class AggregationOperation(Enum):
    # Works with all types, field should be `*`.
    COUNT = "count"
    COUNTIF = "countif"
    COUNT_DISTINCT = "distinct"
    COUNTDISTINCTIF = "distinctif"
    MAKE_SET = "makeset"
    MAKE_SET_IF = "makesetif"

    # Only works for numbers.
    SUM = "sum"
    AVG = "avg"
    MIN = "min"
    MAX = "max"
    TOPK = "topk"
    PERCENTILES = "percentiles"
    HISTOGRAM = "histogram"
    STDEV = "stdev"
    VARIANCE = "variance"
    ARGMIN = "argmin"
    ARGMAX = "argmax"


@dataclass
class Aggregation:
    """Aggregation performed as part of a query."""

    op: AggregationOperation
    field: str = dataclass_field(default="")
    alias: str = dataclass_field(default="")
    argument: any = dataclass_field(default="")



"""This package provides annotation models and methods as well as an AnnotationsClient"""

import ujson
from logging import Logger
from requests import Session
from typing import List, Dict, Optional
from dataclasses import dataclass, asdict, field
from datetime import datetime, timedelta
from urllib.parse import urlencode
from .util import Util


@dataclass
class Annotation:
    """Represents an Axiom annotation"""

    id: str = field(init=False)
    datasets: List[str]
    time: datetime
    endTime: Optional[datetime]
    title: Optional[str]
    description: Optional[str]
    url: Optional[str]
    type: str


@dataclass
class AnnotationCreateRequest:
    """Request used to create an annotation"""

    datasets: List[str]
    time: Optional[datetime]
    endTime: Optional[datetime]
    title: Optional[str]
    description: Optional[str]
    url: Optional[str]
    type: str


@dataclass
class AnnotationUpdateRequest:
    """Request used to update an annotation"""

    datasets: Optional[List[str]]
    time: Optional[datetime]
    endTime: Optional[datetime]
    title: Optional[str]
    description: Optional[str]
    url: Optional[str]
    type: Optional[str]


class AnnotationsClient:  # pylint: disable=R0903
    """AnnotationsClient has methods to manipulate annotations."""

    session: Session

    def __init__(self, session: Session, logger: Logger):
        self.session = session
        self.logger = logger

    def get(self, id: str) -> Annotation:
        """Get a annotation by id."""
        path = "/v2/annotations/%s" % id
        res = self.session.get(path)
        decoded_response = res.json()
        return Util.from_dict(Annotation, decoded_response)

    def create(self, req: AnnotationCreateRequest) -> Annotation:
        """Create an annotation with the given properties."""
        path = "/v2/annotations"
        res = self.session.post(path, data=ujson.dumps(asdict(req)))
        annotation = Util.from_dict(Annotation, res.json())
        self.logger.info(f"created new annotation: {annotation.id}")
        return annotation

    def list(
        self,
        datasets: List[str] = [],
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
    ) -> List[Annotation]:
        """List all annotations."""
        query_params = {}
        if len(datasets) > 0:
            query_params["datasets"] = ",".join(datasets)
        if start != None:
            query_params["start"] = start.isoformat()
        if end != None:
            query_params["end"] = end.isoformat()
        path = f"/v2/annotations?{urlencode(query_params, doseq=True)}"

        res = self.session.get(path)

        annotations = []
        for record in res.json():
            ds = Util.from_dict(Annotation, record)
            annotations.append(ds)

        return annotations

    def update(self, id: str, req: AnnotationUpdateRequest) -> Annotation:
        """Update an annotation with the given properties."""
        path = "/v2/annotations/%s" % id
        res = self.session.put(path, data=ujson.dumps(asdict(req)))
        annotation = Util.from_dict(Annotation, res.json())
        self.logger.info(f"updated annotation({annotation.id})")
        return annotation

    def delete(self, id: str):
        """Deletes an annotation with the given id."""
        path = "/v2/annotations/%s" % id
        self.session.delete(path)



import dacite
import iso8601
from enum import Enum
from uuid import UUID
from typing import Type, TypeVar
from datetime import datetime, timedelta

from .query import QueryKind
from .query.aggregation import AggregationOperation
from .query.result import MessageCode, MessagePriority
from .query.filter import FilterOperation


T = TypeVar("T")


class Util:
    """A collection of helper methods."""

    @classmethod
    def from_dict(cls, data_class: Type[T], data) -> T:
        cfg = dacite.Config(
            type_hooks={
                QueryKind: QueryKind,
                datetime: cls.convert_string_to_datetime,
                AggregationOperation: AggregationOperation,
                FilterOperation: FilterOperation,
                MessageCode: MessageCode,
                MessagePriority: MessagePriority,
                timedelta: cls.convert_string_to_timedelta,
            }
        )

        return dacite.from_dict(data_class=data_class, data=data, config=cfg)

    @classmethod
    def convert_string_to_datetime(cls, val: str) -> datetime:
        d = iso8601.parse_date(val)
        return d

    @classmethod
    def convert_string_to_timedelta(cls, val: str) -> timedelta:
        if val == "0":
            return timedelta(seconds=0)

        exp = "^([0-9]?)([a-z])$"
        import re

        found = re.search(exp, val)
        if not found:
            raise Exception(f"failed to parse timedelta field from value {val}")

        v = int(found.groups()[0])
        unit = found.groups()[1]

        if unit == "s":
            return timedelta(seconds=v)
        elif unit == "m":
            return timedelta(minutes=v)
        elif unit == "h":
            return timedelta(hours=v)
        elif unit == "d":
            return timedelta(days=v)
        else:
            raise Exception(f"failed to parse timedelta field from value {val}")

    @classmethod
    def handle_json_serialization(cls, obj):
        if isinstance(obj, datetime):
            return obj.isoformat("T") + "Z"
        elif isinstance(obj, timedelta):
            return str(obj.seconds) + "s"
        elif isinstance(obj, Enum):
            return obj.value
        elif isinstance(obj, UUID):
            return str(obj)



from dataclasses import dataclass, field
from datetime import datetime, timedelta
from typing import List, Dict, Any, Optional
from enum import Enum

from .query import QueryLegacy


class MessageCode(Enum):
    """Message codes represents the code associated with the query."""

    UNKNOWN_MESSAGE_CODE = ""
    VIRTUAL_FIELD_FINALIZE_ERROR = "virtual_field_finalize_error"
    MISSING_COLUMN = "missing_column"
    LICENSE_LIMIT_FOR_QUERY_WARNING = "license_limit_for_query_warning"
    DEFAULT_LIMIT_WARNING = "default_limit_warning"


class MessagePriority(Enum):
    """Message priorities represents the priority of a message associated with a query."""

    TRACE = "trace"
    DEBUG = "debug"
    INFO = "info"
    WARN = "warn"
    ERROR = "error"
    FATAL = "fatal"


@dataclass
class Message:
    """a message associated with a query result."""

    # priority of the message.
    priority: MessagePriority
    # describes how often a message of this type was raised by the query.
    count: int
    # code of the message.
    code: MessageCode
    # a human readable text representation of the message.
    msg: str


@dataclass
class QueryStatus:
    """the status of a query result."""

    # the duration it took the query to execute.
    elapsedTime: int
    # the amount of blocks that have been examined by the query.
    blocksExamined: int
    # the amount of rows that have been examined by the query.
    rowsExamined: int
    # the amount of rows that matched the query.
    rowsMatched: int
    # the amount of groups returned by the query.
    numGroups: int
    # describes if the query result is a partial result.
    isPartial: bool
    # ContinuationToken is populated when isPartial is true and must be passed
    # to the next query request to retrieve the next result set.
    continuationToken: Optional[str] = field(default=None)
    # describes if the query result is estimated.
    isEstimate: Optional[bool] = field(default=None)
    # the timestamp of the oldest block examined.
    minBlockTime: Optional[str] = field(default=None)
    # the timestamp of the newest block examined.
    maxBlockTime: Optional[str] = field(default=None)
    # messages associated with the query.
    messages: List[Message] = field(default_factory=lambda: [])
    # row id of the newest row, as seen server side.
    maxCursor: Optional[str] = field(default=None)
    # row id of the oldest row, as seen server side.
    minCursor: Optional[str] = field(default=None)


@dataclass
class Entry:
    """Entry is an event that matched a query and is thus part of the result set."""

    # the time the event occurred. Matches SysTime if not specified
    # during ingestion.
    _time: str
    # the time the event was recorded on the server.
    _sysTime: str
    # the unique ID of the event row.
    _rowId: str
    # contains the raw data of the event (with filters and aggregations
    # applied).
    data: Dict[str, Any]


@dataclass
class EntryGroupAgg:
    """an aggregation which is part of a group of queried events."""

    # alias is the aggregations alias. If it wasn't specified at query time, it
    # is the uppercased string representation of the aggregation operation.
    value: Any
    op: str = field(default="")
    # value is the result value of the aggregation.


@dataclass
class EntryGroup:
    """a group of queried event."""

    # the unique id of the group.
    id: int
    # group maps the fieldnames to the unique values for the entry.
    group: Dict[str, Any]
    # aggregations of the group.
    aggregations: List[EntryGroupAgg]


@dataclass
class Interval:
    """the interval of queried time series."""

    # startTime of the interval.
    startTime: datetime
    # endTime of the interval.
    endTime: datetime
    # groups of the interval.
    groups: Optional[List[EntryGroup]]


@dataclass
class Timeseries:
    """Timeseries are queried time series."""

    # the intervals that build a time series.
    series: List[Interval]
    # totals of the time series.
    totals: Optional[List[EntryGroup]]


@dataclass
class QueryLegacyResult:
    """Result is the result of a query."""

    # Status of the query result.
    status: QueryStatus
    # Matches are the events that matched the query.
    matches: List[Entry]
    # Buckets are the time series buckets.
    buckets: Timeseries
    # savedQueryID is the ID of the query that generated this result when it
    # was saved on the server. This is only set when the query was send with
    # the `saveAsKind` option specified.
    savedQueryID: Optional[str] = field(default=None)


@dataclass
class QueryResult:
    """Result is the result of apl query."""

    request: QueryLegacy
    # Status of the apl query result.
    status: QueryStatus
    # Matches are the events that matched the apl query.
    matches: List[Entry]
    # Buckets are the time series buckets.
    buckets: Timeseries
    # Dataset names are the datasets that were used in the apl query.
    dataset_names: List[str] = field(default_factory=lambda: [])
    # savedQueryID is the ID of the apl query that generated this result when it
    # was saved on the server. This is only set when the apl query was send with
    # the `saveAsKind` option specified.
    savedQueryID: Optional[str] = field(default=None)



from enum import Enum
from typing import List
from dataclasses import dataclass, field as dataclass_field


class FilterOperation(Enum):
    """A FilterOperatuib can be applied on queries to filter based on different conditions."""

    EMPTY = ""
    AND = "and"
    OR = "or"
    NOT = "not"

    # Works for strings and numbers.
    EQUAL = "=="
    NOT_EQUAL = "!="
    EXISTS = "exists"
    NOT_EXISTS = "not-exists"

    # Only works for numbers.
    GREATER_THAN = ">"
    GREATER_THAN_EQUAL = ">="
    LESS_THAN = "<"
    LESS_THAN_EQUAL = "<="

    # Only works for strings.
    STARTS_WITH = "starts-with"
    NOT_STARTS_WITH = "not-starts-with"
    ENDS_WITH = "ends-with"
    NOT_ENDS_WITH = "not-ends-with"
    REGEXP = "regexp"
    NOT_REGEXP = "not-regexp"

    # Works for strings and arrays.
    CONTAINS = "contains"
    NOT_CONTAINS = "not-contains"


@dataclass
class BaseFilter:
    op: FilterOperation
    field: str
    value: any
    caseSensitive: bool = dataclass_field(default=False)


@dataclass
class Filter(BaseFilter):
    children: List[BaseFilter] = dataclass_field(default=lambda: [])



from datetime import timedelta
from dataclasses import dataclass, field
from .query import QueryKind


@dataclass
class QueryOptions:
    # StreamingDuration of a query.
    streamingDuration: timedelta = field(default=None)
    # NoCache omits the query cache.
    nocache: bool = field(default=False)
    # SaveKind saves the query on the server with the given query kind. The ID
    # of the saved query is returned with the query result as part of the
    # response. `query.APL` is not a valid kind for this field.
    saveAsKind: QueryKind = field(default=QueryKind.ANALYTICS)
